# This script could be used to obtain all the figures in the Simulation Study 
# of the manuscript by appropriately setting the simulation parameters.
# By default, it provides the top-right panel of Figure 3 of the main text 
# of the manuscript for:
#       - p = 10
#       - budget = 0.2


# Assuming that the working directory is the folder containing this script,
# the following code produces the aforementioned plot using ggplot2,
# which is then stored in the object p, and,
# finally, the figure is saved in a file "plot.pdf"


# This code has been run on computing system with 2 sockets with 24 cores each,
# equipped with Intel(R) Xeon(R) Platinum 8160 processors
# running at a clock frequency of 2.10GHz, and 192 GB of RAM.
# The main information output by the system
# after running the script is reported as follows:

# Resource usage summary:
#   
# CPU time :                                   330959.00 sec.
# Max Memory :                                 9515 MB
# Average Memory :                             7268.14 MB
# Total Requested Memory :                     -
# Delta Memory :                               -
# Max Swap :                                   -
# Max Processes :                              51
# Max Threads :                                53
# Run time :                                   8894 sec.
# Turnaround time :                            113927 sec.


# We also report the output of sessionInfo() obtained from this script as follows:
# R version 3.6.1 (2019-07-05)
# Platform: x86_64-pc-linux-gnu (64-bit)
# Running under: CentOS Linux 7 (Core)
# 
# Matrix products: default
# BLAS:   /afs/.enea.it/software/rproj/R-3.6.1/build/lib64/R/lib/libRblas.so
# LAPACK: /afs/.enea.it/software/rproj/R-3.6.1/build/lib64/R/lib/libRlapack.so
# 
# locale:
# [1] LC_CTYPE=en_US.UTF-8       LC_NUMERIC=C              
# [3] LC_TIME=en_US.UTF-8        LC_COLLATE=en_US.UTF-8    
# [5] LC_MONETARY=en_US.UTF-8    LC_MESSAGES=en_US.UTF-8   
# [7] LC_PAPER=en_US.UTF-8       LC_NAME=C                 
# [9] LC_ADDRESS=C               LC_TELEPHONE=C            
# [11] LC_MEASUREMENT=en_US.UTF-8 LC_IDENTIFICATION=C       
# 
# attached base packages:
# [1] stats     graphics  grDevices utils     datasets  methods   base     
# 
# other attached packages:
# [1] pROC_1.18.5               Rcpp_1.0.4.6             
# [3] RcppArmadillo_0.9.900.1.0 forcats_1.0.0            
# [5] stringr_1.5.0             dplyr_1.1.2              
# [7] purrr_1.0.1               readr_2.1.4              
# [9] tidyr_1.3.0               tibble_3.2.1             
# [11] ggplot2_3.4.2             tidyverse_1.3.0          
# 
# loaded via a namespace (and not attached):
# [1] plyr_1.8.6       pillar_1.9.0     compiler_3.6.1   cellranger_1.1.0
# [5] dbplyr_2.3.2     tools_3.6.1      jsonlite_1.8.4   lubridate_1.9.2 
# [9] lifecycle_1.0.3  gtable_0.3.0     timechange_0.2.0 pkgconfig_2.0.3 
# [13] rlang_1.1.1      reprex_2.0.2     rstudioapi_0.14  DBI_1.1.3       
# [17] cli_3.6.1        parallel_3.6.1   haven_2.5.2      xml2_1.3.4      
# [21] withr_2.5.0      httr_1.4.5       generics_0.1.3   vctrs_0.6.2     
# [25] fs_1.6.2         hms_1.1.3        grid_3.6.1       tidyselect_1.2.0
# [29] glue_1.6.2       R6_2.4.1         fansi_0.4.1      readxl_1.4.2    
# [33] farver_2.0.3     tzdb_0.3.0       modelr_0.1.8     magrittr_2.0.3  
# [37] backports_1.4.1  scales_1.2.1     rvest_1.0.3      colorspace_1.4-1
# [41] labeling_0.3     utf8_1.1.4       stringi_1.7.12   munsell_0.5.0   
# [45] broom_1.0.4      crayon_1.3.4    

# input parameters --------------------------------------------------------------
B <- 0.2
shift_vec <- seq(0, 3, length.out = 6)
p <- 10
weight_entropy <- 0.5

# libs files --------------------------------------------------------------
library(tidyverse)
library(RcppArmadillo)
library(Rcpp)
library(pROC)

source("functions.R")
sourceCpp("funs.cpp")

fun <- function(seed) {
  
  # parameters --------------------------------------------------------------
  cov_matrix <- matrix(0, p, p)
  for (i in 1:p) {
    for (j in 1:p) {
      cov_matrix[i, j] <- 0.75 ^ abs(i - j)
    }
  }
  use_true <- FALSE
  k <- 5
  ntry <- 10
  d1 <- c(shift, rep(0, p - 1))
  d2 <- c(0, shift, rep(0, p - 2))
  TT1 <- 500
  T0 <- 100
  
  # simulate ----------------------------------------------------------------
  set.seed(seed)
  dat <- simulate_hmm_with_constant_length_sequences(
    TT = TT1,
    T_min_IC = 60,
    T_max_IC = 85,
    T_out = 5,
    log_ppi1 = c(0,-Inf),
    log_ppi2 = c(0,-Inf,-Inf),
    logA1 = log(rbind(c(0.96, 0.04), c(0.2, 0.8))),
    logA2 = log(rbind(
      c(0.96, 0.02, 0.02), c(0.2, 0.8, 0), c(0.2, 0, 0.8)
    )),
    means = list(rep(0, p), d1, d2),
    covariances = list(cov_matrix, cov_matrix, cov_matrix)
  )
  x <- c(rep(1, T0), dat$x)
  y <- rbind(mvnfast::rmvn(T0, mu = rep(0, p), sigma = cov_matrix), dat$y)
  TT <- length(x)
  
  
  # MEWMA -------------------------------------------------------------------
  mu0 <- rep(0, p)
  Sigma0 <- cov(y[1:T0, , drop = FALSE])
  center <- colMeans(y[1:T0, , drop = FALSE])
  lambda <- 0.3;
  n <- TT
  
  En <- matrix(nrow = n, ncol = p); 
  V2n <- numeric(n)
  Sigma_En <- lambda / (2 - lambda) * Sigma0; 
  Sigma_En_inv <- solve(Sigma_En)
  En[1, ] <- lambda * y[1, ]
  V2n[1] <- 0
  for (ii in 2:n) {
    if (ii <= T0) {
      ycen <- y[ii, , drop = FALSE] - colMeans(y[1:ii, , drop = FALSE])
      En[ii, ] <- lambda * ycen + (1 - lambda) * En[ii - 1, ]
      V2n[ii] <- 0
    } else {
      cov_robust <- rrcov::CovRobust(y[1:ii, , drop = FALSE])
      ycen <- y[ii, ] - cov_robust$center
      En[ii, ] <- lambda * ycen + (1 - lambda) * En[ii - 1, ]
      V2n[ii] <- En[ii, , drop = FALSE] %*% solve(cov_robust$cov * lambda / (2 - lambda)) %*% En[ii, ]  
    }
    
  }
  lim <- qchisq(0.99, df = p)
  
  
  # unsupervised ------------------------------------------------------------
  print("unsupervised")
  xlabeled_unsup <- x * NA
  xhat_unsupervised <- x * NA
  xhat_unsupervised[1:(T0 + 10)] <- 1
  for (tt in (T0 + 1):length(x)) {
    mod_unsupervised <- baum_welch_R_with_initialization(y = y[1:tt, , drop = FALSE],
                                                         xlabeled = xlabeled_unsup[1:tt],
                                                         equal_covariances = TRUE,
                                                         tol = 1e-3,
                                                         use_true = use_true,
                                                         true_mean =  list(rep(0, p), d1, d2),
                                                         true_covariance = list(cov_matrix),
                                                         k = k,
                                                         ntry = ntry)
    
    xhat_unsupervised[tt] <- which.max(mod_unsupervised$log_gamma[tt, ])
  }


  # random ------------------------------------------------------------------
  print("random")
  which_label_random <- sample((T0 + 1):TT, B * (TT - T0))
  xlabeled_random <- x * NA
  xlabeled_random[which_label_random] <- x[which_label_random]
  xlabeled_random[1:T0] <- 1
  xhat_random <- x * NA
  xhat_random[1:T0] <- 1
  for (tt in (T0 + 1):length(x)) {
    xlabeled_tt <- xlabeled_random[1:tt]
    which_labeled_tt <- which(!is.na(xlabeled_tt))
    nstates_observed <- length(unique(xlabeled_tt[which_labeled_tt]))
    states_observed <- sort(unique(xlabeled_tt[!is.na(xlabeled_tt)]))
    max_nstates <- nstates_observed + sum(is.na(xlabeled_tt))

    if (!are_equal(states_observed, 1:nstates_observed)) {
      actual_states <- states_observed
      actual_xlabeled <- xlabeled_random
      states_observed <- seq_along(states_observed)
      for (ss in 1:nstates_observed) {
        xlabeled_random[xlabeled_random == actual_states[ss]] <- ss
      }
    } else {
      actual_states <- states_observed
    }

    mod_random <- baum_welch_R_with_initialization(y = y[1:tt, , drop = FALSE],
                                                   xlabeled = xlabeled_random[1:tt],
                                                   equal_covariances = TRUE,
                                                   tol = 1e-3,
                                                   use_true = use_true,
                                                   max_nstates = max_nstates,
                                                   true_mean =  list(rep(0, p), d1, d2),
                                                   true_covariance = list(cov_matrix),
                                                   k = k,
                                                   ntry = ntry)
    nstates <- length(mod_random$means_hat)
    if (nstates == 1) {
      chosen_states <- 1
    }
    if (nstates == 2) {
      if (nstates_observed == 1) {
        chosen_states <- 1:2
      }
      if (nstates_observed == 2) {
        chosen_states <- states_observed
      }
    }
    if (nstates == 3) {
      chosen_states <- 1:3
    }
    

    if (!are_equal(actual_states, 1:nstates_observed)) {
      xlabeled_random <- actual_xlabeled
      states_observed <- actual_states
      nstates_observed <- length(states_observed)
    }

    xhat_random[tt] <- chosen_states[which.max(mod_random$log_gamma[tt, ])]

  }


  # equispaced --------------------------------------------------------------
  print("equispaced")
  which_label_equispaced <- as.integer(seq(T0 + 1, length(x), length.out = B * (TT - T0)))
  xlabeled_equispaced <- x * NA
  xlabeled_equispaced[which_label_equispaced] <- x[which_label_equispaced]
  xlabeled_equispaced[1:T0] <- 1
  xhat_equispaced <- x * NA
  xhat_equispaced[1:T0] <- 1
  for (tt in (T0 + 1):length(x)) {
    xlabeled_tt <- xlabeled_equispaced[1:tt]
    which_labeled_tt <- which(!is.na(xlabeled_tt))
    nstates_observed <- length(unique(xlabeled_tt[which_labeled_tt]))
    states_observed <- sort(unique(xlabeled_tt[!is.na(xlabeled_tt)]))
    max_nstates <- nstates_observed + sum(is.na(xlabeled_tt))

    if (!are_equal(states_observed, 1:nstates_observed)) {
      actual_states <- states_observed
      actual_xlabeled <- xlabeled_equispaced
      states_observed <- seq_along(states_observed)
      for (ss in 1:nstates_observed) {
        xlabeled_equispaced[xlabeled_equispaced == actual_states[ss]] <- ss
      }
    } else {
      actual_states <- states_observed
    }

    
    mod_equispaced <- baum_welch_R_with_initialization(y = y[1:tt, , drop = FALSE],
                                                       xlabeled = xlabeled_equispaced[1:tt],
                                                       equal_covariances = TRUE,
                                                       tol = 1e-3,
                                                       use_true = use_true,
                                                       max_nstates = max_nstates,
                                                       true_mean =  list(rep(0, p), d1, d2),
                                                       true_covariance = list(cov_matrix),
                                                       k = k,
                                                       ntry = ntry)
    nstates <- length(mod_equispaced$means_hat)
    if (nstates == 1) {
      chosen_states <- 1
    }
    if (nstates == 2) {
      if (nstates_observed == 1) {
        chosen_states <- 1:2
      }
      if (nstates_observed == 2) {
        chosen_states <- states_observed
      }
    }
    if (nstates == 3) {
      chosen_states <- 1:3
    }

    if (!are_equal(actual_states, 1:nstates_observed)) {
      xlabeled_equispaced <- actual_xlabeled
      states_observed <- actual_states
      nstates_observed <- length(states_observed)
    }

    xhat_equispaced[tt] <- chosen_states[which.max(mod_equispaced$log_gamma[tt, ])]

  }

  
  # stream-based exploration-exploitation -----------------------------------
  print("proposed")
  xlabeled_proposed <- x * NA
  xlabeled_proposed[1:T0] <- x[1:T0]
  log_vec <- rep(NA, length(x))
  entropy_vec <- rep(NA, length(x))
  p_value_exploration_vec <- rep(NA, length(x))
  p_value_exploration_vec_new <- rep(NA, length(x))
  p_value_entropy_vec <- rep(NA, length(x))
  decision <- rep(NA, length(x))
  LRT <- rep(NA, length(x))
  Bvec <- rep(NA, length(x))
  window_length_exploitation <- 1
  window_length_exploration <- 1
  window_length <- max(window_length_exploitation, window_length_exploration)
  available_budget <- B * length(x[-(1:T0)])
  decision[1:(T0+window_length)] <- "1"
  V2n_min <- numeric(TT)
  
  for (ww in (T0 - window_length + 1 + 1):(length(x) - window_length + 1)) {
    ww_end <- (window_length:length(x))[ww]
    ww_start <- ww_end - window_length + 1
    ww_start_exploitation <- ww_end - window_length_exploitation + 1
    ww_start_exploration <- ww_end - window_length_exploration + 1
    
    states_observed <- sort(unique(xlabeled_proposed[!is.na(xlabeled_proposed)]))
    nstates_observed <- length(unique(xlabeled_proposed[!is.na(xlabeled_proposed)]))
    if (!are_equal(states_observed, 1:nstates_observed)) {
      actual_states <- states_observed
      actual_xlabeled <- xlabeled_proposed
      states <- seq_along(states_observed)
      for (ss in 1:nstates_observed) {
        xlabeled_proposed[xlabeled_proposed == actual_states[ss]] <- ss
      }
    } else {
      actual_states <- states_observed
    }
    
    max_nstates <- nstates_observed + sum(is.na(xlabeled_proposed[1:ww_end]))
    
    
    tt <- ww_end
    mod_proposed <- baum_welch_R_with_initialization(y = y[1:ww_end, , drop = FALSE],
                                                     xlabeled = xlabeled_proposed[1:ww_end],
                                                     equal_covariances = TRUE,
                                                     tol = 1e-3,
                                                     use_true = use_true,
                                                     max_nstates = max_nstates,
                                                     true_mean =  list(rep(0, p), d1, d2),
                                                     true_covariance = list(cov_matrix),
                                                     k = k,
                                                     ntry = ntry)
    nstates <- length(mod_proposed$means_hat)
    if (nstates == 1) {
      chosen_states <- 1
    }
    if (nstates == 2) {
      if (nstates_observed == 1) {
        chosen_states <- 1:2  
      }
      if (nstates_observed == 2) {
        chosen_states <- states_observed
      }
    }
    if (nstates == 3) {
      chosen_states <- 1:3
    }
    
    available_budget <- B * length(x[-(1:T0)]) - sum(!is.na(xlabeled_proposed[-(1:T0)]))
    labellable_samples <- length(x) - ww_end + 1
    this_B <- max(min(available_budget / labellable_samples, 1), 0)
    
    nsim_sequences <- max(20, 4 / this_B)
    nsim_sequences <- min(200, nsim_sequences)
    entropy_sim <- numeric(nsim_sequences)
    log_density_sim <- numeric(nsim_sequences)
    which_labeled <- which(!is.na(xlabeled_proposed))
    which_labeled_before_window <- which_labeled[which_labeled < (ww_end - window_length + 1)]
    if (length(which_labeled_before_window) == 0) {
      last_labeled <- 1
      log_ppi <- mod_proposed$log_pi_hat
    } else {
      last_labeled <- max(which_labeled_before_window)
      log_ppi <- rep(-Inf, nstates)
      log_ppi[xlabeled_proposed[last_labeled]] <- 0
    }
    
    window_exploration <- max(which_labeled, ww_start_exploration):ww_end
    nnn <- length(last_labeled:ww_end)
    for (nn in 1:nsim_sequences) {
      dat_sim <- simulate_hmm(nnn,
                              log_ppi = log_ppi,
                              logA = mod_proposed$logAhat,
                              means = mod_proposed$means_hat,
                              covariances = mod_proposed$covariances_hat,
                              xlabeled = xlabeled_proposed[last_labeled:ww_end],
                              n_last_y = nnn)
      entropy_sim[nn] <- entropy_of_sequence_hmm(mod_proposed, 
                                                 xlabeled = xlabeled_proposed[last_labeled:ww_end], 
                                                 y = dat_sim$y, 
                                                 window = (nnn - window_length_exploitation + 1):nnn)
      log_density_sim[nn] <- log_density_of_sequence(mod_proposed,
                                                     xlabeled = xlabeled_proposed[last_labeled:ww_end],
                                                     y = dat_sim$y,
                                                     window = (nnn - length(window_exploration) + 1):nnn)
    }
    entropy_vec[ww_end] <- entropy_of_sequence_hmm(mod_proposed,
                                                   xlabeled = xlabeled_proposed[last_labeled:ww_end],
                                                   y = y[last_labeled:ww_end, , drop = FALSE],
                                                   window = (nnn - window_length_exploitation + 1):nnn)
    log_vec[ww_end] <- log_density_of_sequence(mod_proposed,
                                               xlabeled = xlabeled_proposed[1:ww_end],
                                               y = y[1:ww_end, , drop = FALSE],
                                               window = window_exploration)
    p_value_entropy <- mean(entropy_vec[ww_end] < entropy_sim)
    p_value_entropy_vec[ww_end] <- p_value_entropy
    
    
    p_value_exploration <- mean(log_vec[ww_end] >= log_density_sim)
    p_value_exploration_vec[ww_end] <- p_value_exploration
    
    
    V2n_min[ww_end] <- min_ewma_cpp(y = y, 
                                    mean_list = mod_proposed$means_hat, 
                                    cov_list = mod_proposed$covariances_hat, 
                                    TT = TT, 
                                    p = p, 
                                    lambda = lambda, 
                                    ww_end = ww_end)
    p_value_exploration <- pchisq(V2n_min[ww_end], df = p, lower.tail = FALSE)
    p_value_exploration_vec[ww_end] <- p_value_exploration
    
    
    if (nstates == 1) {
      entropy_vec[ww_end] <- 0
      p_value_entropy <- 1
      p_value_entropy_vec[ww_end] <- p_value_entropy
    }
    
    
    p_value_exploration_new <- 1
    p_value_exploration_vec_new[ww_end] <- p_value_exploration_new
    
    if (!are_equal(actual_states, 1:nstates_observed)) {
      xlabeled_proposed <- actual_xlabeled
      states_observed <- actual_states
      nstates_observed <- length(states_observed)
    }
    
    Bvec[ww_end] <- this_B
    
    Bentropy <- weight_entropy * B
    Bexploration <- (1 - weight_entropy) * B
      
    
    if (p_value_entropy < Bentropy & this_B > B) {
      decision[ww_end] <- "label_exploitation"
      xlabeled_proposed[ww_end] <- x[ww_end]
    } else {
      if (p_value_exploration < Bexploration & this_B > B) {
        decision[ww_end] <- "label_exploration"
        xlabeled_proposed[ww_end] <- x[ww_end]
      } else {
        decision[ww_end] <- chosen_states[which.max(mod_proposed$log_gamma[nrow(mod_proposed$log_gamma), ])]
      }
    }
  }
  
  decision_proposed <- decision
  xhat_proposed <- as.numeric(decision)
  xhat_proposed[!is.na(xlabeled_proposed)] <- xlabeled_proposed[!is.na(xlabeled_proposed)]

  
  
  # results -----------------------------------------------------------------
  xhat_list <- tibble(
    B = B,
    unsupervised = xhat_unsupervised,
    random = xhat_random,
    equispaced = xhat_equispaced,
    proposed = xhat_proposed,
    decision_proposed = decision)
  xlabeled_list <- tibble(
    B = B,
    unsupervised = x * NA,
    random = xlabeled_random,
    equispaced = xlabeled_equispaced,
    proposed = xlabeled_proposed)
  f1_list <- tibble(
    B = B,
    shift = shift,
    p = p,
    weight_entropy = weight_entropy,
    
    unsupervised = get_classification_scores(xhat_unsupervised[-(1:T0)], x[-(1:T0)])$f1,
    random = get_classification_scores(xhat_random[-(1:T0)], x[-(1:T0)])$f1,
    equispaced = get_classification_scores(xhat_equispaced[-(1:T0)], x[-(1:T0)])$f1,
    proposed = get_classification_scores(xhat_proposed[-(1:T0)], x[-(1:T0)])$f1,
    mewma = get_classification_scores(as.numeric(V2n > lim) + 1, as.numeric(x > 1) + 1)$f1,
    
    unsupervised_precision = get_classification_scores(xhat_unsupervised[-(1:T0)], x[-(1:T0)])$precision,
    random_precision = get_classification_scores(xhat_random[-(1:T0)], x[-(1:T0)])$precision,
    equispaced_precision = get_classification_scores(xhat_equispaced[-(1:T0)], x[-(1:T0)])$precision,
    proposed_precision = get_classification_scores(xhat_proposed[-(1:T0)], x[-(1:T0)])$precision,
    mewma_precision = get_classification_scores(as.numeric(V2n > lim) + 1, as.numeric(x > 1) + 1)$precision,
    
    unsupervised_recall = get_classification_scores(xhat_unsupervised[-(1:T0)], x[-(1:T0)])$recall,
    random_recall = get_classification_scores(xhat_random[-(1:T0)], x[-(1:T0)])$recall,
    equispaced_recall = get_classification_scores(xhat_equispaced[-(1:T0)], x[-(1:T0)])$recall,
    proposed_recall = get_classification_scores(xhat_proposed[-(1:T0)], x[-(1:T0)])$recall,
    mewma_recall = get_classification_scores(as.numeric(V2n > lim) + 1, as.numeric(x > 1) + 1)$recall
    
    ) %>% 
    mutate(across(-B, ~ ifelse(is.nan(.x), 0, .x))) %>% 
    as.matrix()
  
  return(f1_list)  
}

nsim <- 48
ncores <- parallel::detectCores(logical = FALSE)
if (.Platform$OS.type == "windows") {
  ncores <- 1
}
out <- list()
for (ss in 1:length(shift_vec)) {
  shift <- shift_vec[ss]
  out[[ss]] <- parallel::mclapply(1:nsim, function(seed) fun(seed), mc.cores = ncores)
  out[[ss]] <- do.call(rbind, out[[ss]])
}
out <- do.call(rbind, out)
# write.table(
#   x = out,
#   file = "output.txt",
#   row.names = TRUE,
#   col.names = TRUE
# )

# plot
methods <- c("proposed", "random", "equispaced", "unsupervised", "MEWMA")
my_colors <- c("black", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd")
my_colors <- setNames(my_colors, methods)

my_lty <- seq_along(methods)
my_lty <- setNames(my_lty, methods)

p <- as.data.frame(out) %>% 
  group_by(B, shift, p) %>%
  select(all_of(tolower(methods))) %>% 
  pivot_longer(all_of(tolower(methods)), names_to = "method", values_to = "f1") %>% 
  group_by(B, shift, p, method) %>%
  summarise(f1_score = mean(f1), sd_f1 = sd(f1) / sqrt(n())) %>% 
  mutate(method = ifelse(method == "mewma", "MEWMA", method)) %>% 
  mutate(p = paste0("p=", p)) %>% 
  mutate(B = paste0("budget=", round(B, 2))) %>% 
  ggplot(aes(shift, f1_score, col = method)) + 
  geom_line(aes(lty = method)) +
  geom_errorbar(aes(ymin = f1_score - 2 * sd_f1, ymax = f1_score + 2 * sd_f1), width = 0.15, show.legend = FALSE) +
  facet_grid(p~B, scales = "free") +
  scale_color_manual(values = my_colors) +
  scale_linetype_manual(values = my_lty) +
  scale_x_continuous(breaks = shift_vec) +
  xlab("shift size") +
  ylab("F1 score") +
  theme_bw() +
  theme(strip.background = element_rect(fill = "white")) +
  theme(aspect.ratio = 1, legend.position = "bottom") +
  theme(legend.key.width = unit(2, "cm"))

ggsave(filename = "plot.pdf", plot = p, width = 10, height = 7)

sessionInfo()
